#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Python operator tables

These are mostly used to resolve the operator in the module operator and to know the list
of operations allowed.

"""

import operator

from nuitka.PythonVersions import python_version

binary_operator_functions = {
    "Add": operator.add,
    "Sub": operator.sub,
    "Pow": operator.pow,
    "Mult": operator.mul,
    "FloorDiv": operator.floordiv,
    "TrueDiv": operator.truediv,
    "Mod": operator.mod,
    "LShift": operator.lshift,
    "RShift": operator.rshift,
    "BitAnd": operator.and_,
    "BitOr": operator.or_,
    "BitXor": operator.xor,
    "Divmod": divmod,
    "IAdd": operator.iadd,
    "ISub": operator.isub,
    "IPow": operator.ipow,
    "IMult": operator.imul,
    "IFloorDiv": operator.ifloordiv,
    "ITrueDiv": operator.itruediv,
    "IMod": operator.imod,
    "ILShift": operator.ilshift,
    "IRShift": operator.irshift,
    "IBitAnd": operator.iand,
    "IBitOr": operator.ior,
    "IBitXor": operator.ixor,
}

# Python 2 only operator
if python_version < 0x300:
    binary_operator_functions["OldDiv"] = operator.div
    binary_operator_functions["IOldDiv"] = operator.idiv

# Python 3.5 only operator
if python_version >= 0x350:
    binary_operator_functions["MatMult"] = operator.matmul
    binary_operator_functions["IMatMult"] = operator.imatmul

unary_operator_functions = {
    "UAdd": operator.pos,
    "USub": operator.neg,
    "Invert": operator.invert,
    "Repr": repr,
    # Boolean not is treated an unary operator.
    "Not": operator.not_,
    "Abs": operator.abs,
}


rich_comparison_functions = {
    "Lt": operator.lt,
    "LtE": operator.le,
    "Eq": operator.eq,
    "NotEq": operator.ne,
    "Gt": operator.gt,
    "GtE": operator.ge,
}

other_comparison_functions = {
    "Is": operator.is_,
    "IsNot": operator.is_not,
    "In": lambda value1, value2: value1 in value2,
    "NotIn": lambda value1, value2: value1 not in value2,
}

comparison_inversions = {
    "Is": "IsNot",
    "IsNot": "Is",
    "In": "NotIn",
    "NotIn": "In",
    "Lt": "GtE",
    "GtE": "Lt",
    "Eq": "NotEq",
    "NotEq": "Eq",
    "Gt": "LtE",
    "LtE": "Gt",
    "exception_match": "exception_mismatch",
    "exception_mismatch": "exception_match",
}

all_comparison_functions = dict(rich_comparison_functions)
all_comparison_functions.update(other_comparison_functions)


def matchException(left, right):
    if python_version >= 0x300:
        if type(right) is tuple:
            for element in right:
                if not isinstance(BaseException, element):
                    raise TypeError(
                        "catching classes that do not inherit from BaseException is not allowed"
                    )
        elif not isinstance(BaseException, right):
            raise TypeError(
                "catching classes that do not inherit from BaseException is not allowed"
            )

    # This doesn't yet work, make it error exit. and silence PyLint for now.
    # pylint: disable=protected-access
    import os

    os._exit(16)

    assert False, left


all_comparison_functions["exception_match"] = matchException



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Standard plug-in to resolve pkg_resource actions at compile time rather than runtime.

Nuitka can detect some things that pkg_resouces may not even be able to during
runtime, e.g. right now checking pip installed versions, is not a thing, while
some packages in their code, e.g. derive their __version__ value from that.
"""


import re

from nuitka.plugins.PluginBase import NuitkaPluginBase


class NuitkaPluginResources(NuitkaPluginBase):
    plugin_name = "pkg-resources"
    plugin_desc = "Resolve version numbers at compile time."

    def __init__(self):
        try:
            import pkg_resources
        except (ImportError, RuntimeError):
            self.pkg_resources = None
        else:
            self.pkg_resources = pkg_resources

        try:
            import importlib_metadata
        except (ImportError, SyntaxError, RuntimeError):
            self.metadata = None
        else:
            self.metadata = importlib_metadata

        # Note: This one is overriding above import, but doesn't need to initialize
        # the value, since it will already be set in case of a problem.
        try:
            from importlib import metadata

            self.metadata = metadata
        except ImportError:
            pass

    @staticmethod
    def isAlwaysEnabled():
        return True

    def onModuleSourceCode(self, module_name, source_code):
        # Many cases to deal with, pylint: disable=too-many-branches

        # This one has strings with false matches, don't attempt those.
        if module_name == "setuptools.command.easy_install":
            return source_code

        if self.pkg_resources:
            for match in re.findall(
                r"""\b(pkg_resources\.get_distribution\(\s*['"](.*?)['"]\s*\)\.((?:parsed_)?version))""",
                source_code,
            ):
                try:
                    value = self.pkg_resources.get_distribution(match[1]).version
                except self.pkg_resources.DistributionNotFound:
                    self.warning(
                        "Cannot find distribution '%s' for '%s', expect potential run time problem."
                        % (match[1], module_name)
                    )
                except Exception:  # catch all, pylint: disable=broad-except
                    self.sysexit(
                        "Error, failed to find distribution '%s', probably a plugin parsing bug for '%s' code."
                        % (match[1], module_name)
                    )
                else:
                    if match[2] == "version":
                        value = repr(value)
                    elif match[2] == "parsed_version":
                        value = (
                            "pkg_resources.extern.packaging.version.Version(%r)" % value
                        )
                    else:
                        assert False

                    source_code = source_code.replace(match[0], value)

            for match in re.findall(
                r"""\b(pkg_resources\.require\(\s*['"](.*?)['"]\s*\))""",
                source_code,
            ):
                # Explicitly call the require function at Nuitka compile time.
                try:
                    self.pkg_resources.require(match[1])
                except self.pkg_resources.ResolutionError:
                    self.warning(
                        "Cannot find requirement '%s' for '%s', expect potential run time problem."
                        % (match[1], module_name)
                    )
                except Exception:  # catch all, pylint: disable=broad-except
                    self.sysexit(
                        "Error, failed to resolve '%s', probably a plugin parsing bug for '%s' code."
                        % (match[1], module_name)
                    )
                else:
                    source_code = source_code.replace(match[0], "")

        if self.metadata:
            for match in re.findall(
                r"""\b((?:importlib_)?metadata\.version\(\s*['"](.*?)['"]\s*\))""",
                source_code,
            ):
                value = self.metadata.version(match[1])
                value = repr(value)

                source_code = source_code.replace(match[0], value)

        return source_code



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Standard plug-in to make multiprocessing work well on Windows.

On Windows, the multiprocessing modules forks new processes which then have
to start from scratch. This won't work if there is no "sys.executable" to
point to a "Python.exe" and won't use compiled code by default.

The issue applies to accelerated and standalone mode alike.
"""

from nuitka import Options
from nuitka.ModuleRegistry import getRootTopModule
from nuitka.plugins.PluginBase import NuitkaPluginBase
from nuitka.PythonVersions import python_version
from nuitka.tree.SourceReading import readSourceCodeFromFilename
from nuitka.utils.ModuleNames import ModuleName


class NuitkaPluginMultiprocessingWorkarounds(NuitkaPluginBase):
    """This is to make multiprocessing work with Nuitka and use compiled code.

    When running in accelerated mode, it's not good to fork a new Python
    instance to run other code, as that won't be accelerated. And when
    run in standalone mode, there may not even be a Python, but it's the
    same principle.

    So by default, this module is on and works around the behavior of the
    "multiprocessing.forking/multiprocessing.spawn/multiprocessing.manager"
    expectations.
    """

    plugin_name = "multiprocessing"
    plugin_desc = "Required by Python's multiprocessing module"

    @classmethod
    def isRelevant(cls):
        return not Options.shallMakeModule()

    @staticmethod
    def getPreprocessorSymbols():
        return {"_NUITKA_PLUGIN_MULTIPROCESSING_ENABLED": "1"}

    @staticmethod
    def createPreModuleLoadCode(module):
        full_name = module.getFullName()

        # TODO: Replace the setting of "sys.frozen" with a change to the source code of the
        # modules we want to affect from this plugin, it's a huge impact on compatibility
        # with other things potentially. We should do it, once the anti-bloat engine is
        # re-usable or supports conditional replacements based on plugin activity and is
        # always on.
        if full_name == "multiprocessing":
            code = """\
import sys, os
sys.frozen = 1
if sys.platform == "win32" and not os.path.exists(sys.argv[0]) and not sys.argv[0].endswith(".exe"):
    sys.executable = sys.argv[0] + ".exe"
else:
    sys.executable = sys.argv[0]
"""
            return (
                code,
                """\
Monkey patching "multiprocessing" load environment.""",
            )

    @staticmethod
    def createPostModuleLoadCode(module):
        full_name = module.getFullName()

        if full_name == "multiprocessing":
            code = """\
try:
    from multiprocessing.forking import ForkingPickler
except ImportError:
    from multiprocessing.reduction import ForkingPickler

class C:
   def f():
       pass

def _reduce_compiled_method(m):
    if m.im_self is None:
        return getattr, (m.im_class, m.im_func.__name__)
    else:
        return getattr, (m.im_self, m.im_func.__name__)

ForkingPickler.register(type(C().f), _reduce_compiled_method)
if str is bytes:
    ForkingPickler.register(type(C.f), _reduce_compiled_method)
"""

            return (
                code,
                """\
Monkey patching "multiprocessing" for compiled methods.""",
            )

    def onModuleInitialSet(self):
        from nuitka.tree.Building import buildModule

        # First, build the module node and then read again from the
        # source code.
        root_module = getRootTopModule()

        module_name = ModuleName("__parents_main__")

        source_code = readSourceCodeFromFilename(module_name, root_module.getFilename())

        # For the call stack, this may look bad or different to what
        # CPython does. Using the "__import__" built-in to not spoil
        # or use the module namespace. The forking module was split up
        # into multiple modules in Python 3.4.
        if python_version >= 0x340:
            source_code += """
__import__("sys").modules["__main__"] = __import__("sys").modules[__name__]
# Not needed, and can crash from minor __file__ differences, depending on invocation
__import__("multiprocessing.spawn").spawn._fixup_main_from_path = lambda mod_name : None
__import__("multiprocessing.spawn").spawn.freeze_support()"""
        else:
            source_code += """
__import__("sys").modules["__main__"] = __import__("sys").modules[__name__]
__import__("multiprocessing.forking").forking.freeze_support()"""

        multiprocessing_main_module, _added = buildModule(
            module_filename=root_module.getCompileTimeFilename(),
            module_package=module_name.getPackageName(),
            source_code=source_code,
            is_top=False,
            is_main=False,
            is_shlib=False,
            is_fake=module_name,
            hide_syntax_error=False,
        )

        if multiprocessing_main_module.getCompilationMode() == "bytecode":
            multiprocessing_main_module.setSourceCode(source_code)

        yield multiprocessing_main_module

    def onModuleEncounter(self, module_filename, module_name, module_kind):
        # Enforce recursion in to multiprocessing for accelerated mode, which
        # would normally avoid this.
        if module_name.hasNamespace("multiprocessing"):
            return True, "Multiprocessing plugin needs this to monkey patch it."

    def decideCompilation(self, module_name, source_ref):
        if module_name.hasNamespace("multiprocessing"):
            return "bytecode"

        # TODO: Make this demotable too.
        # or module_name in( "multiprocessing-preLoad", "multiprocessing-postLoad"):

    def onFrozenModuleSourceCode(self, module_name, is_package, source_code):
        if module_name == "multiprocessing.resource_tracker":
            source_code = source_code.replace(
                "args += ['-c', cmd % r]",
                "args += ['--multiprocessing-resource-tracker', str(r)]",
            )

        return source_code


class NuitkaPluginDetectorMultiprocessingWorkarounds(NuitkaPluginBase):
    detector_for = NuitkaPluginMultiprocessingWorkarounds

    @classmethod
    def isRelevant(cls):
        return not Options.shallMakeModule()

    def checkModuleSourceCode(self, module_name, source_code):
        if module_name == "__main__":
            if "multiprocessing" in source_code:
                self.warnUnusedPlugin("Multiprocessing workarounds for compiled code.")



#     Copyright 2021, Jorj McKie, mailto:<jorj.x.mckie@outlook.de>
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Details see below in class definition.
"""
import os
import re
import shutil
import sys
from collections import namedtuple

from nuitka import Options
from nuitka.freezer.IncludedEntryPoints import makeDllEntryPoint
from nuitka.plugins.PluginBase import NuitkaPluginBase
from nuitka.plugins.Plugins import getActiveQtPlugin, hasActivePlugin
from nuitka.PythonVersions import getSystemPrefixPath
from nuitka.utils import Execution
from nuitka.utils.FileOperations import (
    getFileList,
    listDir,
    makePath,
    putTextFileContents,
)
from nuitka.utils.Utils import getOS, isWin32Windows


def getNumpyCoreBinaries(module):
    """Return any binaries in numpy/core and/or numpy/.libs.

    Notes:
        This covers the special cases like MKL binaries.

    Returns:
        tuple of abspaths of binaries.
    """
    numpy_dir = module.getCompileTimeDirectory()
    numpy_core_dir = os.path.join(numpy_dir, "core")
    base_prefix = getSystemPrefixPath()

    binaries = []

    # first look in numpy/.libs for binaries
    libdir = os.path.join(numpy_dir, ".libs" if getOS() != "Darwin" else ".dylibs")
    suffix_start = len(libdir) + 1
    if os.path.isdir(libdir):
        dlls_pkg = os.listdir(libdir)
        binaries += [[os.path.join(libdir, f), suffix_start] for f in dlls_pkg]

    # then look for libraries in numpy.core package path
    # should already return the MKL files in ordinary cases

    re_anylib = re.compile(r"\w+\.(?:dll|so|dylib)", re.IGNORECASE)

    dlls_pkg = [f for f in os.listdir(numpy_core_dir) if re_anylib.match(f)]
    binaries += [[os.path.join(numpy_core_dir, f), suffix_start] for f in dlls_pkg]

    # Also look for MKL libraries in folder "above" numpy.
    # This should meet the layout of Anaconda installs.

    if isWin32Windows():
        lib_dir = os.path.join(base_prefix, "Library", "bin")
        suffix_start = len(lib_dir) + 1
    else:
        lib_dir = os.path.join(base_prefix, "lib")
        suffix_start = len(lib_dir) + 1

    if not os.path.isdir(lib_dir):
        return binaries

    re_mkllib = re.compile(r"^(?:lib)?mkl\w+\.(?:dll|so|dylib)", re.IGNORECASE)

    for f in os.listdir(lib_dir):
        if isWin32Windows():
            if not (f.startswith(("libi", "libm", "mkl")) and f.endswith(".dll")):
                continue
        else:
            if not re_mkllib.match(f):
                continue

        binaries.append([os.path.join(lib_dir, f), suffix_start])

    return binaries


class NumpyPlugin(NuitkaPluginBase):
    """This class represents the main logic of the plugin.

    This is a plugin to ensure scripts using numpy, scipy, matplotlib, pandas,
    scikit-learn, etc. work well in standalone mode.

    While there already are relevant entries in the "ImplicitImports.py" plugin,
    this plugin copies any additional binary or data files required by many
    installations.

    """

    plugin_name = "numpy"  # Nuitka knows us by this name
    plugin_desc = "Required for numpy, scipy, pandas, matplotlib, etc."

    def __init__(self, include_matplotlib, include_scipy):
        self.matplotlib = include_matplotlib
        self.scipy = include_scipy

        self.enabled_plugins = None  # list of active standard plugins
        self.numpy_copied = False  # indicator: numpy files copied
        self.scipy_copied = True  # indicator: scipy files copied
        if self.scipy:
            self.scipy_copied = False

        self.mpl_data_copied = False  # indicator: matplotlib data copied

        # Information about matplotlib install.
        self.matplotlib_info = None

    @classmethod
    def isRelevant(cls):
        """Check whether plugin might be required.

        Returns:
            True if this is a standalone compilation.
        """
        return Options.isStandaloneMode()

    @classmethod
    def addPluginCommandLineOptions(cls, group):
        group.add_option(
            "--noinclude-scipy",
            action="store_false",
            dest="include_scipy",
            default=True,
            help="""\
Should scipy, sklearn or skimage when used be not included with numpy, Default is %default.""",
        )

        group.add_option(
            "--noinclude-matplotlib",
            action="store_false",
            dest="include_matplotlib",
            default=True,
            help="""\
Should matplotlib not be be included with numpy, Default is %default.""",
        )

    def considerExtraDlls(self, dist_dir, module):
        """Copy extra shared libraries or data for this installation.

        Args:
            dist_dir: the name of the program's dist folder
            module: module object
        Returns:
            empty tuple
        """
        full_name = module.getFullName()

        if not self.numpy_copied and full_name == "numpy":
            self.numpy_copied = True
            binaries = getNumpyCoreBinaries(module)

            for f in binaries:
                bin_file, idx = f  # (filename, pos. prefix + 1)
                back_end = bin_file[idx:]
                tar_file = os.path.join(dist_dir, back_end)
                makePath(  # create any missing intermediate folders
                    os.path.dirname(tar_file)
                )
                shutil.copyfile(bin_file, tar_file)

            bin_total = len(binaries)  # anything there at all?
            if bin_total > 0:
                msg = "Copied %i %s from 'numpy' installation." % (
                    bin_total,
                    "file" if bin_total < 2 else "files",
                )
                self.info(msg)

        if os.name == "nt" and not self.scipy_copied and full_name == "scipy":
            # TODO: We are not getting called twice, are we?
            assert not self.scipy_copied
            self.scipy_copied = True

            bin_total = 0
            for entry_point in self._getScipyCoreBinaries(
                scipy_dir=module.getCompileTimeDirectory()
            ):
                yield entry_point
                bin_total += 1

            if bin_total > 0:
                msg = "Copied %i %s from 'scipy' installation." % (
                    bin_total,
                    "file" if bin_total < 2 else "files",
                )
                self.info(msg)

        # TODO: Ouch, do not copy data files when asked to copy DLLs.
        if self.matplotlib and full_name == "matplotlib" and not self.mpl_data_copied:
            self.mpl_data_copied = True

            self.copyMplDataFiles(dist_dir)

    def _getMatplotlibInfo(self):
        """Determine the filename of matplotlibrc and the default backend, etc.

        Notes:
            There might exist a local version outside 'matplotlib/mpl-data' which
            we then must use instead. Determine its name by aksing matplotlib.
        """
        # TODO: Replace this with using self.queryRuntimeInformationMultiple to remove
        # code duplication.
        if self.matplotlib_info is None:
            cmd = r"""\
from __future__ import print_function
from matplotlib import matplotlib_fname, get_backend, __version__
try:
    from matplotlib import get_data_path
except ImportError:
    from matplotlib import _get_data_path as get_data_path
from inspect import getsource
print(repr(matplotlib_fname()))
print(repr(get_backend()))
print(repr(get_data_path()))
print(repr(__version__))
print(repr("MATPLOTLIBDATA" in getsource(get_data_path)))
"""

            # TODO: Make this is a re-usable pattern, output from a script with values per line
            feedback = Execution.check_output([sys.executable, "-c", cmd])

            if str is not bytes:  # ensure str in Py3 and up
                feedback = feedback.decode("utf8")

            # Ignore Windows newlines difference.
            feedback = feedback.replace("\r", "")

            MatplotlibInfo = namedtuple(
                "MatplotlibInfo",
                (
                    "matplotlibrc_filename",
                    "backend",
                    "data_path",
                    "matplotlib_version",
                    "needs_matplotlibdata_env",
                ),
            )

            # We are being lazy here, the code is trusted, pylint: disable=eval-used
            self.matplotlib_info = MatplotlibInfo(
                *(eval(value) for value in feedback.splitlines())
            )

        return self.matplotlib_info

    def copyMplDataFiles(self, dist_dir):
        """Write matplotlib data files ('mpl-data')."""

        matplotlib_info = self._getMatplotlibInfo()

        if not os.path.isdir(matplotlib_info.data_path):
            self.sysexit(
                "mpl-data missing, matplotlib installation appears to be broken"
            )

        # Copy data files to dist folder
        for fullname in getFileList(matplotlib_info.data_path):
            filename = os.path.relpath(fullname, matplotlib_info.data_path)

            if filename.endswith("matplotlibrc"):  # handle config separately
                continue

            target_filename = os.path.join(dist_dir, "matplotlib", "mpl-data", filename)

            makePath(os.path.dirname(target_filename))  # create intermediate folders
            shutil.copyfile(fullname, target_filename)

        old_lines = (
            open(matplotlib_info.matplotlibrc_filename).read().splitlines()
        )  # old config file lines
        new_lines = []  # new config file lines

        found = False  # checks whether backend definition encountered
        for line in old_lines:
            line = line.rstrip()

            # omit meaningless lines
            if line.startswith("#") and matplotlib_info.matplotlib_version < "3":
                continue

            new_lines.append(line)

            if line.startswith(("backend ", "backend:")):
                # old config file has a backend definition
                found = True

        if not found and matplotlib_info.matplotlib_version < "3":
            # Set the backend, so even if it was run time determined, we now enforce it.
            new_lines.append("backend: %s" % matplotlib_info.backend)

        matplotlibrc_filename = os.path.join(
            dist_dir, "matplotlib", "mpl-data", "matplotlibrc"
        )

        putTextFileContents(filename=matplotlibrc_filename, contents=new_lines)

        self.info("Copied 'matplotlib/mpl-data'.")

    @staticmethod
    def _getScipyCoreBinaries(scipy_dir):
        """Return binaries from the extra-dlls folder (Windows only)."""

        for dll_dir_name in ("extra_dll", ".libs"):
            dll_dir_path = os.path.join(scipy_dir, dll_dir_name)

            if os.path.isdir(dll_dir_path):
                for source_path, source_filename in listDir(dll_dir_path):
                    if source_filename.lower().endswith(".dll"):
                        yield makeDllEntryPoint(
                            source_path=source_path,
                            dest_path=os.path.join(
                                "scipy", dll_dir_name, source_filename
                            ),
                            package_name="scipy",
                        )

    def onModuleEncounter(self, module_filename, module_name, module_kind):
        # pylint: disable=too-many-branches,too-many-return-statements
        if not self.scipy and module_name.hasOneOfNamespaces(
            "scipy", "sklearn", "skimage"
        ):
            return False, "Omit unneeded components"

        if not self.matplotlib and module_name.hasOneOfNamespaces(
            "matplotlib", "skimage"
        ):
            return False, "Omit unneeded components"

        if module_name == "scipy.sparse.csgraph._validation":
            return True, "Replicate implicit import"

        if self.matplotlib and module_name.hasNamespace("mpl_toolkits"):
            return True, "Needed by matplotlib"

        if module_name in ("cv2", "cv2.cv2", "cv2.data"):
            return True, "Needed for OpenCV"

        sklearn_mods = [
            "sklearn.utils.sparsetools._graph_validation",
            "sklearn.utils.sparsetools._graph_tools",
            "sklearn.utils.lgamma",
            "sklearn.utils.weight_vector",
            "sklearn.utils._unittest_backport",
            "sklearn.externals.joblib.externals.cloudpickle.dumps",
            "sklearn.externals.joblib.externals.loky.backend.managers",
        ]

        if isWin32Windows():
            sklearn_mods.extend(
                [
                    "sklearn.externals.joblib.externals.loky.backend.synchronize",
                    "sklearn.externals.joblib.externals.loky.backend._win_wait",
                    "sklearn.externals.joblib.externals.loky.backend._win_reduction",
                    "sklearn.externals.joblib.externals.loky.backend.popen_loky_win32",
                ]
            )
        else:
            sklearn_mods.extend(
                [
                    "sklearn.externals.joblib.externals.loky.backend.synchronize",
                    "sklearn.externals.joblib.externals.loky.backend.compat_posix",
                    "sklearn.externals.joblib.externals.loky.backend._posix_reduction",
                    "sklearn.externals.joblib.externals.loky.backend.popen_loky_posix",
                ]
            )

        if self.scipy and module_name in sklearn_mods:
            return True, "Needed by sklearn"

        # some special handling for matplotlib:
        # depending on whether 'tk-inter' resp. 'qt-plugins' are enabled,
        # matplotlib backends are included.
        if self.matplotlib:
            if hasActivePlugin("tk-inter"):
                if module_name in (
                    "matplotlib.backends.backend_tk",
                    "matplotlib.backends.backend_tkagg",
                    "matplotlib.backend.tkagg",
                ):
                    return True, "Needed for tkinter matplotplib backend"

            if getActiveQtPlugin() is not None:
                # Note, their code tries everything behind that name, the qt5 is
                # misleading therefore, PySide will work there too.
                if module_name in (
                    "matplotlib.backends.backend_qt5",
                    "matplotlib.backends.backend_qt5.py",
                    "matplotlib.backends.backend_qt5cairo.py",
                    "matplotlib.backend.backend_qt5.py",
                ):
                    return True, "Needed for Qt matplotplib backend"

            if module_name == "matplotlib.backends.backend_agg":
                return True, "Needed as standard matplotplib backend"

    def createPreModuleLoadCode(self, module):
        """Method called when a module is being imported.

        Notes:
            If full name equals "matplotlib" we insert code to set the
            environment variable that e.g. Debian versions of matplotlib
            use.

        Args:
            module: the module object
        Returns:
            Code to insert and descriptive text (tuple), or (None, None).
        """

        # Matplotlib might be off, or the version may not need the environment variable.
        if (
            self.matplotlib
            and module.getFullName() == "matplotlib"
            and self._getMatplotlibInfo().needs_matplotlibdata_env
        ):
            code = r"""
import os
os.environ["MATPLOTLIBDATA"] = os.path.join(__nuitka_binary_dir, "matplotlib", "mpl-data")
"""
            return (
                code,
                "Setting 'MATPLOTLIBDATA' environment variable for matplotlib to find package data.",
            )


class NumpyPluginDetector(NuitkaPluginBase):
    """Only used if plugin is NOT activated.

    Notes:
        We are given the chance to issue a warning if we think we may be required.
    """

    detector_for = NumpyPlugin

    @classmethod
    def isRelevant(cls):
        """Check whether plugin might be required.

        Returns:
            True if this is a standalone compilation.
        """
        return Options.isStandaloneMode()

    def onModuleDiscovered(self, module):
        """This method checks whether numpy is required.

        Notes:
            For this we check whether its first name part is numpy relevant.
        Args:
            module: the module object
        Returns:
            None
        """
        module_name = module.getFullName()
        if module_name.hasOneOfNamespaces(
            "numpy", "scipy", "skimage", "pandas", "matplotlib", "sklearn"
        ):
            self.warnUnusedPlugin(
                "Numpy support for at least '%s'."
                % module_name.getTopLevelPackageName()
            )



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Codes for classes.

Most the class specific stuff is solved in re-formulation. Only the selection
of the metaclass remains as specific.
"""

from nuitka.PythonVersions import python_version

from .CodeHelpers import (
    generateChildExpressionsCode,
    withObjectCodeTemporaryAssignment,
)
from .ErrorCodes import getErrorExitCode


def generateSelectMetaclassCode(to_name, expression, emit, context):
    metaclass_name, bases_name = generateChildExpressionsCode(
        expression=expression, emit=emit, context=context
    )

    # This is used for Python3 only.
    assert python_version >= 0x300

    arg_names = [metaclass_name, bases_name]

    with withObjectCodeTemporaryAssignment(
        to_name, "metaclass_result", expression, emit, context
    ) as value_name:

        emit(
            "%s = SELECT_METACLASS(%s);"
            % (value_name, ", ".join(str(arg_name) for arg_name in arg_names))
        )

        getErrorExitCode(
            check_name=value_name, release_names=arg_names, emit=emit, context=context
        )

        context.addCleanupTempName(value_name)


def generateBuiltinSuperCode(to_name, expression, emit, context):
    type_name, object_name = generateChildExpressionsCode(
        expression=expression, emit=emit, context=context
    )

    with withObjectCodeTemporaryAssignment(
        to_name, "super_value", expression, emit, context
    ) as value_name:

        emit(
            "%s = BUILTIN_SUPER%d(%s, %s);"
            % (
                value_name,
                2 if expression.isExpressionBuiltinSuper2() else 0,
                type_name if type_name is not None else "NULL",
                object_name if object_name is not None else "NULL",
            )
        )

        getErrorExitCode(
            check_name=value_name,
            release_names=(type_name, object_name),
            emit=emit,
            context=context,
        )

        context.addCleanupTempName(value_name)



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Spawning processes.

This is to replace the standard spawn implementation with one that tracks the
progress, and gives warnings about things taking very long.
"""

import os
import subprocess
import sys
import threading

from nuitka.Tracing import my_print, scons_logger
from nuitka.utils.Timing import TimerReport

from .SconsCaching import runClCache
from .SconsProgress import closeSconsProgressBar, updateSconsProgressBar
from .SconsUtils import decodeData


# Thread class to run a command
class SubprocessThread(threading.Thread):
    def __init__(self, cmdline, env):
        threading.Thread.__init__(self)

        self.cmdline = cmdline
        self.env = env

        self.data = None
        self.err = None
        self.exit_code = None
        self.exception = None

        self.timer_report = TimerReport(
            message="Running %s took %%.2f seconds"
            % repr(self.cmdline).replace("%", "%%"),
            min_report_time=60,
            logger=scons_logger,
        )

    def run(self):
        try:
            # execute the command, queue the result
            with self.timer_report:
                proc = subprocess.Popen(
                    self.cmdline,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=False,
                    env=self.env,
                )

                self.data, self.err = proc.communicate()
                self.exit_code = proc.wait()
        except Exception as e:  # will rethrow all, pylint: disable=broad-except
            self.exception = e

    def getProcessResult(self):
        return self.data, self.err, self.exit_code, self.exception


def runProcessMonitored(cmdline, env):
    thread = SubprocessThread(cmdline, env)
    thread.start()

    # Allow a minute before warning for long compile time.
    thread.join(60)

    if thread.is_alive():
        scons_logger.info(
            "Slow C compilation detected, used %.0fs so far, this might indicate scalability problems."
            % thread.timer_report.getTimer().getDelta()
        )

    thread.join()

    updateSconsProgressBar()

    return thread.getProcessResult()


def _filterLinkOutput(module_mode, lto_mode, data, exit_code):
    # Training newline in some cases, esp. LTO it seems.
    data = data.rstrip()

    if module_mode:
        data = b"\r\n".join(
            line
            for line in data.split(b"\r\n")
            if b"   Creating library" not in line
            # On localized compilers, the message to ignore is not as clear.
            if not (module_mode and b".exp" in line)
        )

    # The linker will say generating code at the end, due to localization
    # we don't know.
    if lto_mode and exit_code == 0:
        if len(data.split(b"\r\n")) == 2:
            data = b""

    return data


# To work around Windows not supporting command lines of greater than 10K by
# default:
def getWindowsSpawnFunction(module_mode, lto_mode, source_files):
    def spawnWindowsCommand(
        sh, escape, cmd, args, env
    ):  # pylint: disable=unused-argument

        # The "del" appears to not work reliably, but is used with large amounts of
        # files to link. So, lets do this ourselves, plus it avoids a process
        # spawn.
        if cmd == "del":
            assert len(args) == 2

            os.unlink(args[1])
            return 0

        # For quoted arguments that end in a backslash, things don't work well
        # this is a workaround for it.
        def removeTrailingSlashQuote(arg):
            if arg.endswith(r"\""):
                return arg[:-1] + '\\"'
            else:
                return arg

        newargs = " ".join(removeTrailingSlashQuote(arg) for arg in args[1:])
        cmdline = cmd + " " + newargs

        # Special hook for clcache inline copy
        if cmd == "<clcache>":
            data, err, rv = runClCache(args, env)
        else:
            data, err, rv, exception = runProcessMonitored(cmdline, env)

            if exception:
                closeSconsProgressBar()
                raise exception

        if cmd == "link":
            data = _filterLinkOutput(
                module_mode=module_mode, lto_mode=lto_mode, data=data, exit_code=rv
            )

        elif cmd in ("cl", "<clcache>"):
            # Skip forced output from cl.exe
            data = data[data.find(b"\r\n") + 2 :]

            source_basenames = [
                os.path.basename(source_file) for source_file in source_files
            ]

            def check(line):
                return line in (b"", b"Generating Code...") or line in source_basenames

            data = (
                b"\r\n".join(line for line in data.split(b"\r\n") if not check(line))
                + b"\r\n"
            )

        if data is not None and data.rstrip():
            my_print("Unexpected output from this command:", style="yellow")
            my_print(cmdline, style="yellow")

            if str is not bytes:
                data = decodeData(data)

            my_print(data, style="yellow", end="")

        if err:
            if str is not bytes:
                err = decodeData(err)

            my_print(err, style="yellow", end="")

        return rv

    return spawnWindowsCommand


def _unescape(arg):
    # Undo the damage that scons did to pass it to "sh"
    arg = arg.strip('"')

    slash = "\\"
    special = '"$()'

    arg = arg.replace(slash + slash, slash)
    for c in special:
        arg = arg.replace(slash + c, c)

    return arg


def isIgnoredError(line):
    # Many cases, pylint: disable=too-many-return-statements

    # Debian Python2 static libpython lto warnings:
    if b"function `posix_tmpnam':" in line:
        return True
    if b"function `posix_tempnam':" in line:
        return True

    # Self compiled Python2 static libpython lot warnings:
    if b"the use of `tmpnam_r' is dangerous" in line:
        return True
    if b"the use of `tempnam' is dangerous" in line:
        return True
    if line.startswith((b"Objects/structseq.c:", b"Python/import.c:")):
        return True
    if line == b"In function 'load_next',":
        return True
    if b"at Python/import.c" in line:
        return True

    # Bullseys when compiling in directory with spaces:
    if b"overriding recipe for target" in line:
        return True
    if b"ignoring old recipe for target" in line:
        return True
    if b"Error 1 (ignored)" in line:
        return True

    return False


def subprocess_spawn(args):
    sh, _cmd, args, env = args

    proc = subprocess.Popen(
        [sh, "-c", " ".join(args)], env=env, close_fds=True, stderr=subprocess.PIPE
    )

    _stdout, stderr = proc.communicate()

    stderr_lines = stderr.splitlines()

    ignore_next = False
    for line in stderr_lines:
        if ignore_next:
            ignore_next = False
            continue

        if isIgnoredError(line):
            ignore_next = True
            continue

        if str is not bytes:
            line = decodeData(line)

        my_print(line, style="yellow", file=sys.stderr)

    return proc.wait()


class SpawnThread(threading.Thread):
    def __init__(self, *args):
        threading.Thread.__init__(self)

        self.args = args

        self.timer_report = TimerReport(
            message="Running %s took %%.2f seconds"
            % (" ".join(_unescape(arg) for arg in self.args[2]).replace("%", "%%"),),
            min_report_time=60,
            logger=scons_logger,
        )

        self.result = None
        self.exception = None

    def run(self):
        try:
            # execute the command, queue the result
            with self.timer_report:
                self.result = subprocess_spawn(self.args)
        except Exception as e:  # will rethrow all, pylint: disable=broad-except
            self.exception = e

    def getSpawnResult(self):
        return self.result, self.exception


def runSpawnMonitored(sh, cmd, args, env):
    thread = SpawnThread(sh, cmd, args, env)
    thread.start()

    # Allow a minute before warning for long compile time.
    thread.join(60)

    if thread.is_alive():
        scons_logger.info(
            "Slow C compilation detected, used %.0fs so far, this might indicate scalability problems."
            % thread.timer_report.getTimer().getDelta()
        )

    thread.join()

    updateSconsProgressBar()

    return thread.getSpawnResult()


def getWrappedSpawnFunction():
    def spawnCommand(sh, escape, cmd, args, env):
        # signature needed towards Scons core, pylint: disable=unused-argument

        # Avoid using ccache on binary constants blob, not useful and not working
        # with old ccache.
        if '"__constants_data.o"' in args or '"__constants_data.os"' in args:
            env = dict(env)
            env["CCACHE_DISABLE"] = "1"

        result, exception = runSpawnMonitored(sh, cmd, args, env)

        if exception:
            closeSconsProgressBar()

            raise exception

        return result

    return spawnCommand


def enableSpawnMonitoring(env, win_target, module_mode, lto_mode, source_files):
    if win_target:
        env["SPAWN"] = getWindowsSpawnFunction(
            module_mode=module_mode, lto_mode=lto_mode, source_files=source_files
        )
    else:
        env["SPAWN"] = getWrappedSpawnFunction()



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" For macOS application bundle creation

"""
import os
import shutil

from nuitka import Options, OutputDirectories

from .FileOperations import makePath, openTextFile
from .Images import convertImageToIconFormat


def createPlistInfoFile(logger, onefile):
    # Many details, pylint: disable=too-many-locals

    import plistlib

    if Options.isStandaloneMode():
        bundle_dir = os.path.dirname(OutputDirectories.getStandaloneDirectoryPath())
    else:
        bundle_dir = os.path.dirname(
            OutputDirectories.getResultRunFilename(onefile=onefile)
        )

    result_filename = OutputDirectories.getResultFullpath(onefile=onefile)
    app_name = Options.getMacOSAppName() or os.path.basename(result_filename)

    signed_app_name = Options.getMacOSSignedAppName() or app_name

    # TODO: We want an OrderedDict probably for stability.
    infos = {
        "CFBundleDisplayName": app_name,
        "CFBundleName": app_name,
        "CFBundleIdentifier": signed_app_name,
        "CFBundleExecutable": app_name,
        "CFBundleInfoDictionaryVersion": "6.0",
        "CFBundlePackageType": "APPL",
        # TODO: Make this an option too.
        "CFBundleShortVersionString": "1.0.0",
    }

    icon_paths = Options.getIconPaths()

    if icon_paths:
        assert len(icon_paths) == 1
        icon_path = icon_paths[0]

        # Convert to single macOS .icns file if necessary
        if not icon_path.endswith(".icns"):
            logger.info(
                "File '%s' is not in macOS icon format, converting to it." % icon_path
            )

            icon_build_path = os.path.join(
                OutputDirectories.getSourceDirectoryPath(onefile=onefile),
                "icons",
            )
            makePath(icon_build_path)
            converted_icon_path = os.path.join(
                icon_build_path,
                "Icons.icns",
            )

            convertImageToIconFormat(
                logger=logger,
                image_filename=icon_path,
                icon_filename=converted_icon_path,
            )
            icon_path = converted_icon_path

        icon_name = os.path.basename(icon_path)
        resources_dir = os.path.join(bundle_dir, "Resources")
        makePath(resources_dir)

        shutil.copyfile(icon_path, os.path.join(resources_dir, icon_name))

        infos["CFBundleIconFile"] = icon_name

    # Console mode, which is why we have to use bundle in the first place typically.
    if Options.shallDisableConsoleWindow():
        infos["NSHighResolutionCapable"] = True
    else:
        infos["LSBackgroundOnly"] = True

    filename = os.path.join(bundle_dir, "Info.plist")

    if str is bytes:
        plist_contents = plistlib.writePlistToString(infos)
    else:
        plist_contents = plistlib.dumps(infos)

    with openTextFile(filename=filename, mode="wb") as plist_file:
        plist_file.write(plist_contents)



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Shared definitions of what helpers are available.

These are functions to work with helper names, as well as sets of functions to
generate specialized code variants for.

Note: These are ordered, so we can define the order they are created in the
code generation of specialized helpers, as this set is used for input there
too.

"""

from nuitka.containers.oset import OrderedSet


def parseTypesFromHelper(helper_name):
    """Function to parse helper names."""

    if "_INPLACE" not in helper_name:
        target_code = helper_name.split("_")[3]
        left_code = helper_name.split("_")[4]
        right_code = helper_name.split("_")[5]
    else:
        target_code = None
        left_code = helper_name.split("_")[3]
        right_code = helper_name.split("_")[4]

    return target_code, left_code, right_code


def deriveInplaceFromBinaryOperations(operations_set):
    """Derive the in-place operations from the binary ones.

    These can largely be the same, or should be, and keeping them inline is easier when
    generating them. Obviously the templates might both need changes to optimize equally
    well for all variants.
    """

    if not operations_set:
        return None

    operation = next(iter(operations_set)).split("_")[2]

    return OrderedSet(
        helper_name.replace(operation + "_OBJECT", operation) + "_INPLACE"
        for helper_name in operations_set
        if parseTypesFromHelper(helper_name)[0] == "OBJECT"
        if "CLONG"
        not in parseTypesFromHelper(
            helper_name
        )  # TODO: In-place template must be enhanced for these.
    )


def buildOrderedSet(*producers):
    values = []

    for producer in producers:
        values.extend(producer)

    return OrderedSet(values)


def makeTypeOps(op_code, type_name, include_nbool):
    yield "BINARY_OPERATION_%s_OBJECT_%s_%s" % (op_code, type_name, type_name)
    yield "BINARY_OPERATION_%s_OBJECT_OBJECT_%s" % (op_code, type_name)
    yield "BINARY_OPERATION_%s_OBJECT_%s_OBJECT" % (op_code, type_name)

    if include_nbool:
        for helper in makeTypeOpsNbool(op_code, type_name):
            yield helper


def makeTypeOpsNbool(op_code, type_name):
    yield "BINARY_OPERATION_%s_NBOOL_%s_%s" % (op_code, type_name, type_name)
    yield "BINARY_OPERATION_%s_NBOOL_OBJECT_%s" % (op_code, type_name)
    yield "BINARY_OPERATION_%s_NBOOL_%s_OBJECT" % (op_code, type_name)


def makeFriendOps(op_code, include_nbool, *type_names):
    for type_name1 in type_names:
        for type_name2 in type_names:
            if type_name1 == type_name2:
                continue

            yield "BINARY_OPERATION_%s_OBJECT_%s_%s" % (op_code, type_name1, type_name2)
            yield "BINARY_OPERATION_%s_OBJECT_%s_%s" % (op_code, type_name2, type_name1)

            if include_nbool:
                yield "BINARY_OPERATION_%s_NBOOL_%s_%s" % (
                    op_code,
                    type_name1,
                    type_name2,
                )
                yield "BINARY_OPERATION_%s_NBOOL_%s_%s" % (
                    op_code,
                    type_name2,
                    type_name1,
                )


def makeDefaultOps(op_code, include_nbool):
    yield "BINARY_OPERATION_%s_OBJECT_OBJECT_OBJECT" % op_code
    if include_nbool:
        yield "BINARY_OPERATION_%s_NBOOL_OBJECT_OBJECT" % op_code


def makeNonContainerMathOps(op_code):
    for type_name in "TUPLE", "LIST", "DICT", "SET", "FROZENSET":
        if "BIT" in op_code and type_name == "SET":
            continue
        if "SUB" in op_code and type_name == "SET":
            continue

        for value in makeTypeOps(op_code, type_name, include_nbool=True):
            yield value


specialized_add_helpers_set = buildOrderedSet(
    makeTypeOps("ADD", "INT", include_nbool=True),
    makeTypeOps("ADD", "LONG", include_nbool=True),
    makeTypeOps("ADD", "FLOAT", include_nbool=True),
    makeTypeOps("ADD", "STR", include_nbool=False),
    makeTypeOps("ADD", "UNICODE", include_nbool=False),
    makeTypeOps("ADD", "BYTES", include_nbool=False),
    makeTypeOps("ADD", "TUPLE", include_nbool=False),
    makeTypeOps("ADD", "LIST", include_nbool=True),
    # These are friends naturally, they all add with another
    makeFriendOps("ADD", True, "INT", "LONG", "FLOAT"),
    # TODO: Make CLONG ready to join above group.
    # makeFriendOps("ADD", True, "INT", "CLONG"),
    # These are friends too.
    makeFriendOps("ADD", True, "STR", "UNICODE"),
    # Default implementation.
    makeDefaultOps("ADD", include_nbool=True),
)

nonspecialized_add_helpers_set = buildOrderedSet(
    makeTypeOpsNbool("ADD", "STR"),  # Not really likely to be used.
    makeTypeOpsNbool("ADD", "UNICODE"),  # Not really likely to be used.
    makeTypeOpsNbool("ADD", "BYTES"),  # Not really likely to be used.
    makeTypeOpsNbool("ADD", "TUPLE"),  # Not really likely to be used.
)

specialized_sub_helpers_set = buildOrderedSet(
    makeTypeOps("SUB", "INT", include_nbool=True),
    makeTypeOps("SUB", "LONG", include_nbool=True),
    makeTypeOps("SUB", "FLOAT", include_nbool=True),
    # These are friends naturally, they all sub with another
    makeFriendOps("SUB", True, "INT", "LONG", "FLOAT"),
    makeDefaultOps("SUB", include_nbool=True),
)

# These made no sense to specialize for, nothing to gain.
nonspecialized_sub_helpers_set = buildOrderedSet(
    makeTypeOps("SUB", "STR", include_nbool=True),
    makeTypeOps("SUB", "UNICODE", include_nbool=True),
    makeTypeOps("SUB", "BYTES", include_nbool=True),
    makeNonContainerMathOps("SUB"),
)

specialized_mult_helpers_set = buildOrderedSet(
    makeTypeOps("MULT", "INT", include_nbool=True),
    makeTypeOps("MULT", "LONG", include_nbool=True),
    makeTypeOps("MULT", "FLOAT", include_nbool=True),
    (
        "BINARY_OPERATION_MULT_OBJECT_CLONG_CLONG",
        "BINARY_OPERATION_MULT_OBJECT_INT_CLONG",
        "BINARY_OPERATION_MULT_OBJECT_CLONG_INT",
        #        "BINARY_OPERATION_MULT_OBJECT_LONG_CLONG",
        #        "BINARY_OPERATION_MULT_OBJECT_CLONG_LONG",
        "BINARY_OPERATION_MULT_OBJECT_OBJECT_STR",
        "BINARY_OPERATION_MULT_OBJECT_STR_OBJECT",
        "BINARY_OPERATION_MULT_OBJECT_INT_STR",
        "BINARY_OPERATION_MULT_OBJECT_STR_INT",
        "BINARY_OPERATION_MULT_OBJECT_LONG_STR",
        "BINARY_OPERATION_MULT_OBJECT_STR_LONG",
        "BINARY_OPERATION_MULT_OBJECT_OBJECT_UNICODE",
        "BINARY_OPERATION_MULT_OBJECT_UNICODE_OBJECT",
        "BINARY_OPERATION_MULT_OBJECT_INT_UNICODE",
        "BINARY_OPERATION_MULT_OBJECT_UNICODE_INT",
        "BINARY_OPERATION_MULT_OBJECT_LONG_UNICODE",
        "BINARY_OPERATION_MULT_OBJECT_UNICODE_LONG",
        "BINARY_OPERATION_MULT_OBJECT_OBJECT_TUPLE",
        "BINARY_OPERATION_MULT_OBJECT_TUPLE_OBJECT",
        "BINARY_OPERATION_MULT_OBJECT_INT_TUPLE",
        "BINARY_OPERATION_MULT_OBJECT_TUPLE_INT",
        "BINARY_OPERATION_MULT_OBJECT_LONG_TUPLE",
        "BINARY_OPERATION_MULT_OBJECT_TUPLE_LONG",
        "BINARY_OPERATION_MULT_OBJECT_OBJECT_LIST",
        "BINARY_OPERATION_MULT_OBJECT_LIST_OBJECT",
        "BINARY_OPERATION_MULT_OBJECT_INT_LIST",
        "BINARY_OPERATION_MULT_OBJECT_LIST_INT",
        "BINARY_OPERATION_MULT_OBJECT_LONG_LIST",
        "BINARY_OPERATION_MULT_OBJECT_LIST_LONG",
        "BINARY_OPERATION_MULT_OBJECT_OBJECT_BYTES",
        "BINARY_OPERATION_MULT_OBJECT_BYTES_OBJECT",
        "BINARY_OPERATION_MULT_OBJECT_LONG_BYTES",
        "BINARY_OPERATION_MULT_OBJECT_BYTES_LONG",
    ),
    # These are friends naturally, they all mul with another
    makeFriendOps("MULT", True, "INT", "LONG", "FLOAT"),
    makeDefaultOps("MULT", include_nbool=True),
)

nonspecialized_mult_helpers_set = None

specialized_truediv_helpers_set = buildOrderedSet(
    makeTypeOps("TRUEDIV", "INT", include_nbool=True),
    makeTypeOps("TRUEDIV", "LONG", include_nbool=True),
    makeTypeOps("TRUEDIV", "FLOAT", include_nbool=True),
    # These are friends naturally, they div mul with another
    makeFriendOps("TRUEDIV", True, "INT", "LONG", "FLOAT"),
    makeDefaultOps("TRUEDIV", include_nbool=True),
)

nonspecialized_truediv_helpers_set = buildOrderedSet(
    makeTypeOps("TRUEDIV", "UNICODE", include_nbool=True),
    makeTypeOps("TRUEDIV", "STR", include_nbool=True),
    makeTypeOps("TRUEDIV", "BYTES", include_nbool=True),
    makeNonContainerMathOps("TRUEDIV"),
)

specialized_olddiv_helpers_set = OrderedSet(
    helper.replace("TRUEDIV", "OLDDIV") for helper in specialized_truediv_helpers_set
)

nonspecialized_olddiv_helpers_set = OrderedSet(
    helper.replace("TRUEDIV", "OLDDIV") for helper in nonspecialized_truediv_helpers_set
)

specialized_floordiv_helpers_set = OrderedSet(
    helper.replace("TRUEDIV", "FLOORDIV") for helper in specialized_truediv_helpers_set
)

nonspecialized_floordiv_helpers_set = OrderedSet(
    helper.replace("TRUEDIV", "FLOORDIV")
    for helper in nonspecialized_truediv_helpers_set
)

specialized_mod_helpers_set = buildOrderedSet(
    makeTypeOps("MOD", "INT", include_nbool=True),
    makeTypeOps("MOD", "LONG", include_nbool=True),
    makeTypeOps("MOD", "FLOAT", include_nbool=True),
    # These are friends naturally, they mod with another
    makeFriendOps("MOD", True, "INT", "LONG", "FLOAT"),
    (
        # String interpolation with STR:
        "BINARY_OPERATION_MOD_OBJECT_STR_INT",
        "BINARY_OPERATION_MOD_OBJECT_STR_LONG",
        "BINARY_OPERATION_MOD_OBJECT_STR_FLOAT",
        "BINARY_OPERATION_MOD_OBJECT_STR_STR",
        "BINARY_OPERATION_MOD_OBJECT_STR_UNICODE",
        "BINARY_OPERATION_MOD_OBJECT_STR_TUPLE",
        "BINARY_OPERATION_MOD_OBJECT_STR_LIST",
        "BINARY_OPERATION_MOD_OBJECT_STR_DICT",
        "BINARY_OPERATION_MOD_OBJECT_STR_OBJECT",
        # String formatting with UNICODE:
        "BINARY_OPERATION_MOD_OBJECT_UNICODE_INT",
        "BINARY_OPERATION_MOD_OBJECT_UNICODE_LONG",
        "BINARY_OPERATION_MOD_OBJECT_UNICODE_FLOAT",
        "BINARY_OPERATION_MOD_OBJECT_UNICODE_STR",
        "BINARY_OPERATION_MOD_OBJECT_UNICODE_BYTES",
        "BINARY_OPERATION_MOD_OBJECT_UNICODE_UNICODE",
        "BINARY_OPERATION_MOD_OBJECT_UNICODE_TUPLE",
        "BINARY_OPERATION_MOD_OBJECT_UNICODE_LIST",
        "BINARY_OPERATION_MOD_OBJECT_UNICODE_DICT",
        "BINARY_OPERATION_MOD_OBJECT_UNICODE_OBJECT",
        # String formatting with BYTES:
        "BINARY_OPERATION_MOD_OBJECT_BYTES_BYTES",
        "BINARY_OPERATION_MOD_OBJECT_BYTES_LONG",
        "BINARY_OPERATION_MOD_OBJECT_BYTES_FLOAT",
        "BINARY_OPERATION_MOD_OBJECT_BYTES_UNICODE",
        "BINARY_OPERATION_MOD_OBJECT_BYTES_TUPLE",
        "BINARY_OPERATION_MOD_OBJECT_BYTES_LIST",
        "BINARY_OPERATION_MOD_OBJECT_BYTES_DICT",
        "BINARY_OPERATION_MOD_OBJECT_BYTES_OBJECT",
        # String formatting with OBJECT:
        "BINARY_OPERATION_MOD_OBJECT_OBJECT_STR",
        "BINARY_OPERATION_MOD_OBJECT_OBJECT_BYTES",
        "BINARY_OPERATION_MOD_OBJECT_OBJECT_UNICODE",
        "BINARY_OPERATION_MOD_OBJECT_OBJECT_TUPLE",
        "BINARY_OPERATION_MOD_OBJECT_OBJECT_LIST",
        "BINARY_OPERATION_MOD_OBJECT_OBJECT_DICT",
        # Default implementation.
        "BINARY_OPERATION_MOD_OBJECT_OBJECT_OBJECT",
    ),
    makeDefaultOps("MOD", include_nbool=True),
)

nonspecialized_mod_helpers_set = buildOrderedSet(
    (
        "BINARY_OPERATION_MOD_OBJECT_TUPLE_OBJECT",
        "BINARY_OPERATION_MOD_OBJECT_LIST_OBJECT",
        # String formatting with STR:
        "BINARY_OPERATION_MOD_NBOOL_STR_INT",
        "BINARY_OPERATION_MOD_NBOOL_STR_LONG",
        "BINARY_OPERATION_MOD_NBOOL_STR_FLOAT",
        "BINARY_OPERATION_MOD_NBOOL_STR_STR",
        "BINARY_OPERATION_MOD_NBOOL_STR_UNICODE",
        "BINARY_OPERATION_MOD_NBOOL_STR_TUPLE",
        "BINARY_OPERATION_MOD_NBOOL_STR_LIST",
        "BINARY_OPERATION_MOD_NBOOL_STR_DICT",
        "BINARY_OPERATION_MOD_NBOOL_STR_OBJECT",
        # String formatting with UNICODE:
        "BINARY_OPERATION_MOD_NBOOL_UNICODE_INT",
        "BINARY_OPERATION_MOD_NBOOL_UNICODE_LONG",
        "BINARY_OPERATION_MOD_NBOOL_UNICODE_FLOAT",
        "BINARY_OPERATION_MOD_NBOOL_UNICODE_STR",
        "BINARY_OPERATION_MOD_NBOOL_UNICODE_BYTES",
        "BINARY_OPERATION_MOD_NBOOL_UNICODE_UNICODE",
        "BINARY_OPERATION_MOD_NBOOL_UNICODE_TUPLE",
        "BINARY_OPERATION_MOD_NBOOL_UNICODE_LIST",
        "BINARY_OPERATION_MOD_NBOOL_UNICODE_DICT",
        "BINARY_OPERATION_MOD_NBOOL_UNICODE_OBJECT",
    )
)

specialized_imod_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_mod_helpers_set
)

nonspecialized_imod_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_mod_helpers_set
)


specialized_bitor_helpers_set = buildOrderedSet(
    makeTypeOps("BITOR", "LONG", include_nbool=True),
    makeTypeOps("BITOR", "INT", include_nbool=True),
    makeFriendOps("BITOR", True, "INT", "LONG"),
    (
        # Set containers can do this
        "BINARY_OPERATION_BITOR_OBJECT_SET_SET",
        "BINARY_OPERATION_BITOR_OBJECT_OBJECT_SET",
        "BINARY_OPERATION_BITOR_OBJECT_SET_OBJECT",
    ),
    makeDefaultOps("BITOR", include_nbool=True),
)
nonspecialized_bitor_helpers_set = buildOrderedSet(
    makeTypeOps("BITOR", "FLOAT", include_nbool=True),
    makeNonContainerMathOps("BITOR"),
    makeTypeOps("BITOR", "UNICODE", include_nbool=True),
    makeTypeOps("BITOR", "STR", include_nbool=True),
    makeTypeOps("BITOR", "BYTES", include_nbool=True),
)

specialized_bitand_helpers_set = OrderedSet(
    helper.replace("_BITOR_", "_BITAND_") for helper in specialized_bitor_helpers_set
)
nonspecialized_bitand_helpers_set = OrderedSet(
    helper.replace("_BITOR_", "_BITAND_") for helper in nonspecialized_bitor_helpers_set
)
specialized_bitxor_helpers_set = OrderedSet(
    helper.replace("_BITOR_", "_BITXOR_") for helper in specialized_bitor_helpers_set
)
nonspecialized_bitxor_helpers_set = OrderedSet(
    helper.replace("_BITOR_", "_BITXOR_") for helper in nonspecialized_bitor_helpers_set
)
specialized_lshift_helpers_set = OrderedSet(
    helper.replace("_BITOR_", "_LSHIFT_")
    for helper in specialized_bitor_helpers_set
    if "_SET" not in helper
    if "_TUPLE" not in helper
)
nonspecialized_lshift_helpers_set = buildOrderedSet(
    makeTypeOps("BITOR", "FLOAT", include_nbool=True), makeNonContainerMathOps("BITOR")
)
nonspecialized_lshift_helpers_set = OrderedSet(
    helper.replace("_BITOR_", "_LSHIFT_") for helper in nonspecialized_bitor_helpers_set
)
specialized_rshift_helpers_set = OrderedSet(
    helper.replace("_LSHIFT_", "_RSHIFT_") for helper in specialized_lshift_helpers_set
)
nonspecialized_rshift_helpers_set = OrderedSet(
    helper.replace("_LSHIFT_", "_RSHIFT_")
    for helper in nonspecialized_lshift_helpers_set
)


specialized_pow_helpers_set = OrderedSet(
    (
        # Float is used by other types for ** operations.
        "BINARY_OPERATION_POW_OBJECT_FLOAT_FLOAT",
        "BINARY_OPERATION_POW_NBOOL_FLOAT_FLOAT",
        "BINARY_OPERATION_POW_OBJECT_OBJECT_FLOAT",
        "BINARY_OPERATION_POW_NBOOL_OBJECT_FLOAT",
        "BINARY_OPERATION_POW_OBJECT_FLOAT_OBJECT",
        "BINARY_OPERATION_POW_NBOOL_FLOAT_OBJECT",
        "BINARY_OPERATION_POW_OBJECT_LONG_LONG",
        "BINARY_OPERATION_POW_NBOOL_LONG_LONG",
        "BINARY_OPERATION_POW_OBJECT_OBJECT_LONG",
        "BINARY_OPERATION_POW_NBOOL_OBJECT_LONG",
        "BINARY_OPERATION_POW_OBJECT_LONG_OBJECT",
        "BINARY_OPERATION_POW_NBOOL_LONG_OBJECT",
        # TODO: Enable these later.
        #        "BINARY_OPERATION_POW_OBJECT_LONG_FLOAT",
        #        "BINARY_OPERATION_POW_NBOOL_LONG_FLOAT",
        "BINARY_OPERATION_POW_OBJECT_INT_INT",
        "BINARY_OPERATION_POW_NBOOL_INT_INT",
        "BINARY_OPERATION_POW_OBJECT_OBJECT_INT",
        "BINARY_OPERATION_POW_NBOOL_OBJECT_INT",
        "BINARY_OPERATION_POW_OBJECT_INT_OBJECT",
        "BINARY_OPERATION_POW_NBOOL_INT_OBJECT",
        "BINARY_OPERATION_POW_OBJECT_LONG_INT",
        "BINARY_OPERATION_POW_NBOOL_LONG_INT",
        "BINARY_OPERATION_POW_OBJECT_INT_LONG",
        "BINARY_OPERATION_POW_NBOOL_INT_LONG",
        # Default implementation.
        "BINARY_OPERATION_POW_OBJECT_OBJECT_OBJECT",
        "BINARY_OPERATION_POW_NBOOL_OBJECT_OBJECT",
    )
)
nonspecialized_pow_helpers_set = buildOrderedSet(
    makeTypeOps("POW", "STR", include_nbool=True),
    makeTypeOps("POW", "UNICODE", include_nbool=True),
    makeTypeOps("POW", "BYTES", include_nbool=True),
    makeNonContainerMathOps("POW"),
)


specialized_divmod_helpers_set = buildOrderedSet(
    makeTypeOps("DIVMOD", "INT", include_nbool=False),
    makeTypeOps("DIVMOD", "LONG", include_nbool=False),
    makeTypeOps("DIVMOD", "FLOAT", include_nbool=False),
    # These are friends naturally, they mod with another
    # makeFriendOps("DIVMOD", False, "INT", "LONG", "FLOAT"),
    makeDefaultOps("DIVMOD", include_nbool=False),
)
nonspecialized_divmod_helpers_set = buildOrderedSet(
    makeTypeOpsNbool("DIVMOD", "INT"),
    makeTypeOpsNbool("DIVMOD", "LONG"),
    makeTypeOpsNbool("DIVMOD", "FLOAT"),
    makeTypeOps("DIVMOD", "UNICODE", include_nbool=True),
    makeTypeOps("DIVMOD", "STR", include_nbool=True),
    makeTypeOps("DIVMOD", "BYTES", include_nbool=True),
    makeNonContainerMathOps("DIVMOD"),
)

assert (
    "BINARY_OPERATION_DIVMOD_OBJECT_TUPLE_OBJECT" in nonspecialized_divmod_helpers_set
)

specialized_matmult_helpers_set = buildOrderedSet(
    makeTypeOps("MATMULT", "LONG", include_nbool=False),
    makeTypeOps("MATMULT", "FLOAT", include_nbool=False),
    makeDefaultOps("MATMULT", include_nbool=False),
)

nonspecialized_matmult_helpers_set = buildOrderedSet(
    makeTypeOpsNbool("MATMULT", "LONG"),
    makeTypeOpsNbool("MATMULT", "FLOAT"),
    makeTypeOps("MATMULT", "TUPLE", include_nbool=True),
    makeTypeOps("MATMULT", "LIST", include_nbool=True),
    makeTypeOps("MATMULT", "DICT", include_nbool=True),
    makeTypeOps("MATMULT", "BYTES", include_nbool=True),
    makeTypeOps("MATMULT", "UNICODE", include_nbool=True),
)

specialized_iadd_helpers_set = buildOrderedSet(
    deriveInplaceFromBinaryOperations(specialized_add_helpers_set),
    ("BINARY_OPERATION_ADD_LIST_TUPLE_INPLACE",),
)
nonspecialized_iadd_helpers_set = buildOrderedSet(
    deriveInplaceFromBinaryOperations(nonspecialized_add_helpers_set),
    (
        "BINARY_OPERATION_ADD_LIST_STR_INPLACE",
        "BINARY_OPERATION_ADD_LIST_BYTES_INPLACE",
        "BINARY_OPERATION_ADD_LIST_BYTEARRAY_INPLACE",
        "BINARY_OPERATION_ADD_LIST_UNICODE_INPLACE",
        "BINARY_OPERATION_ADD_LIST_SET_INPLACE",  # semi useful only
        "BINARY_OPERATION_ADD_LIST_FROZENSET_INPLACE",
        "BINARY_OPERATION_ADD_LIST_DICT_INPLACE",
    ),
)

specialized_isub_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_sub_helpers_set
)
nonspecialized_isub_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_sub_helpers_set
)

specialized_imult_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_mult_helpers_set
)

nonspecialized_imult_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_mult_helpers_set
)

specialized_ibitor_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_bitor_helpers_set
)

nonspecialized_ibitor_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_bitor_helpers_set
)

specialized_ibitand_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_bitand_helpers_set
)

nonspecialized_ibitand_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_bitand_helpers_set
)

specialized_ibitxor_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_bitxor_helpers_set
)

nonspecialized_ibitxor_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_bitxor_helpers_set
)

specialized_ilshift_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_lshift_helpers_set
)

nonspecialized_ilshift_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_lshift_helpers_set
)

specialized_irshift_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_rshift_helpers_set
)

nonspecialized_irshift_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_rshift_helpers_set
)

specialized_ifloordiv_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_floordiv_helpers_set
)

nonspecialized_ifloordiv_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_floordiv_helpers_set
)

specialized_itruediv_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_truediv_helpers_set
)

nonspecialized_itruediv_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_truediv_helpers_set
)

specialized_iolddiv_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_olddiv_helpers_set
)

nonspecialized_iolddiv_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_olddiv_helpers_set
)

specialized_ipow_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_pow_helpers_set
)

nonspecialized_ipow_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_pow_helpers_set
)

specialized_imatmult_helpers_set = deriveInplaceFromBinaryOperations(
    specialized_matmult_helpers_set
)

nonspecialized_imatmult_helpers_set = deriveInplaceFromBinaryOperations(
    nonspecialized_matmult_helpers_set
)


def getSpecializedOperations(operator):
    return globals()["specialized_%s_helpers_set" % operator.lower()]


def getNonSpecializedOperations(operator):
    return globals()["nonspecialized_%s_helpers_set" % operator.lower()]


def getCodeNameForOperation(operator):
    return operator[1:].upper() if operator[0] == "I" else operator.upper()



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Common tests for built-ins.

Attempting to cover generic run time behavior. Static optimization of these is
not expected.
"""


def yield_helper():
    print("Yielding 40")
    yield 40
    print("Yielding 60")
    yield 60
    print("Yielding 30")
    yield 30


class CommonOptimizationTest:
    _list_tests = {
        "Test on empty list": [],
        "Test on all None": [None, None, None],
        "Test on single True and rest None": [None, True, None],
        "Test on zero list 20000 times": [0] * 20000,
        "Test on zero list 255 times": [0] * 255,
    }

    _range_tests = {
        "Test on range with single argument": range(260),
        "Test on range with two arguments": range(1, 270),
        "Test on range with three arguments": range(2, 1024, 5),
    }

    _yeild_tests = {"Yield Test": yield_helper()}

    _non_iterable_tests = {
        "Float value": 1.0,
        "Type value": float,
        "Int value": 1,
        "Complex value": 1j,
    }

    _set_tests = {"Test on set": set([0, 1, 2, 3, 3])}

    _dict_tests = {"Test on dict": {1: "One", 2: "Two"}}

    _other_tests = {
        "Test for string": "String",
        "Test for unicode": u"Unicode",
        "Test for byte": b"byte",
    }

    def __init__(self, builtin):
        self.builtin = builtin

    def print_tests(self, test):
        for desc, value in sorted(test.items()):
            try:
                print("{}: {}".format(desc, self.builtin(value)))
            except Exception as e:  # pylint: disable=broad-except
                print("caught ", repr(e))

    def run_all_tests(self):
        print("Calling without args:")
        try:
            print(self.builtin())
        except Exception as e:  # pylint: disable=broad-except
            print("caught ", repr(e))

        tests = [
            self._list_tests,
            self._range_tests,
            self._set_tests,
            self._dict_tests,
            self._yeild_tests,
            self._non_iterable_tests,
            self._other_tests,
        ]

        for test in tests:
            self.print_tests(test)


# TODO: Right now this tests NOTHING of Nuitka, only the Python built-in
print("Test for any")
any_test = CommonOptimizationTest(any)
any_test.run_all_tests()

print("Test for all")
all_test = CommonOptimizationTest(all)
all_test.run_all_tests()



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Signing of executables.

"""

from nuitka.Tracing import postprocessing_logger

from .Execution import executeToolChecked
from .FileOperations import withMadeWritableFileMode

_macos_codesign_usage = "The 'codesign' is used to remove invalidated signatures on macOS and required to be found."


def removeMacOSCodeSignature(filename):
    """Remove the code signature from a filename.

    Args:
        filename - The file to be modified.

    Returns:
        None

    Notes:
        This is macOS specific.
    """

    with withMadeWritableFileMode(filename):
        executeToolChecked(
            logger=postprocessing_logger,
            command=["codesign", "--remove-signature", "--all-architectures", filename],
            absence_message=_macos_codesign_usage,
        )


def addMacOSCodeSignature(filename, identity, entitlements_filename, deep):
    extra_args = []

    # Weak signing is supported.
    if not identity:
        identity = "-"

    command = [
        "codesign",
        "-s",
        identity,
        "--force",
        "--timestamp",
        "--all-architectures",
    ]

    # hardened runtime unless no good identify
    if identity != "-":
        extra_args.append("--options=runtime")

    if entitlements_filename:
        extra_args.append("--entitlements")
        extra_args.append(entitlements_filename)

    if deep:
        extra_args.append("--deep")

    command.append(filename)

    with withMadeWritableFileMode(filename):
        executeToolChecked(
            logger=postprocessing_logger,
            command=command,
            absence_message=_macos_codesign_usage,
        )



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Utility module.

Here the small things that fit nowhere else and don't deserve their own module.

"""

import os
import platform
import sys

from nuitka.PythonVersions import python_version


def getOS():
    if os.name == "nt":
        return "Windows"
    elif os.name == "posix":
        result = os.uname()[0]

        # Handle msys2 posix nature still meaning it's Windows.
        if result.startswith(("MSYS_NT-", "MINGW64_NT-")):
            result = "Windows"

        return result
    else:
        assert False, os.name


_linux_distribution = None


def getLinuxDistribution():
    """Name of the Linux distribution.

    We should usually avoid this, and rather test for the feature,
    but in some cases it's hard to manage that.
    """
    # singleton, pylint: disable=global-statement
    global _linux_distribution

    if getOS() != "Linux":
        return None

    if _linux_distribution is None:
        # pylint: disable=I0021,deprecated-method,no-member
        try:
            result = platform.dist()[0].title()
        except AttributeError:
            from .Execution import check_output

            try:
                result = check_output(["lsb_release", "-i", "-s"], shell=False).title()

                if str is not bytes:
                    result = result.decode("utf8")
            except FileNotFoundError:
                from .FileOperations import getFileContentByLine

                for line in getFileContentByLine("/etc/os-release"):
                    if line.startswith("ID="):
                        result = line[3:]
                        break
                else:
                    from nuitka.Tracing import general

                    general.sysexit("Error, cannot detect Linux distribution.")

        _linux_distribution = result.title()

    return _linux_distribution


def isDebianBasedLinux():
    # TODO: What is with Mint, maybe others, this list should be expanded potentially.
    return getLinuxDistribution() in ("Debian", "Ubuntu")


def isWin32Windows():
    """The Win32 variants of Python does have win32 only, not posix."""
    return os.name == "nt"


def isPosixWindows():
    """The MSYS2 variant of Python does have posix only, not Win32."""
    return os.name == "posix" and getOS() == "Windows"


_is_alpine = None


def isAlpineLinux():
    if os.name == "posix":

        # Avoid repeated file system lookup, pylint: disable=global-statement
        global _is_alpine
        if _is_alpine is None:
            _is_alpine = os.path.isfile("/etc/alpine-release")

        return _is_alpine
    else:
        return False


def getArchitecture():
    if getOS() == "Windows":
        if "AMD64" in sys.version:
            return "x86_64"
        else:
            return "x86"
    else:
        return os.uname()[4]


def getCoreCount():
    cpu_count = 0

    # Try to sum up the CPU cores, if the kernel shows them.
    try:
        # Try to get the number of logical processors
        with open("/proc/cpuinfo") as cpuinfo_file:
            cpu_count = cpuinfo_file.read().count("processor\t:")
    except IOError:
        pass

    if not cpu_count:
        import multiprocessing

        cpu_count = multiprocessing.cpu_count()

    return cpu_count


def encodeNonAscii(var_name):
    """Encode variable name that is potentially not ASCII to ASCII only.

    For Python3, unicode identifiers can be used, but these are not
    possible in C, so we need to replace them.
    """
    if python_version < 0x300:
        return var_name
    else:
        # Using a escaping here, because that makes it safe in terms of not
        # to occur in the encoding escape sequence for unicode use.
        var_name = var_name.replace("$$", "$_$")

        var_name = var_name.encode("ascii", "xmlcharrefreplace")
        var_name = var_name.decode("ascii")

        return var_name.replace("&#", "$$").replace(";", "")


def hasOnefileSupportedOS():
    return getOS() in ("Linux", "Windows", "Darwin", "FreeBSD")


def getUserName():
    """Return the user name.

    Notes: Currently doesn't work on Windows.
    """
    import pwd  # pylint: disable=I0021,import-error

    return pwd.getpwuid(os.getuid())[0]



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
def g():
    for a in range(3):
        yield a

    return 7


def h():
    yield 4
    yield 5


def f():
    print("Yielded from returner", (yield g()))
    print("Yielded from non-return value", (yield h()))


print("Result", list(f()))

print("Yielder with return value", list(g()))

# This will raise when looking up any attribute.
class Broken:
    def __iter__(self):
        return self

    def __next__(self):
        return 1

    def __getattr__(self, attr):
        1 / 0


def test_broken_getattr_handling():
    def g():
        yield from Broken()

    print("Next with send: ", end="")
    try:
        gi = g()
        next(gi)
        gi.send(1)
    except Exception as e:
        print("Caught", repr(e))

    print("Next with throw: ", end="")
    try:
        gi = g()
        next(gi)
        gi.throw(AttributeError)
    except Exception as e:
        print("Caught", repr(e))

    print("Next with close: ", end="")
    try:
        gi = g()
        next(gi)
        gi.close()

        print("All good")
    except Exception as e:
        print("Caught", repr(e))


test_broken_getattr_handling()


def test_throw_catched_subgenerator_handling():
    def g1():
        try:
            print("Starting g1")
            yield "g1 ham"
            yield from g2()
            yield "g1 eggs"
        finally:
            print("Finishing g1")

    def g2():
        try:
            print("Starting g2")
            yield "g2 spam"
            yield "g2 more spam"
        except LunchError:
            print("Caught LunchError in g2")
            yield "g2 lunch saved"
            yield "g2 yet more spam"

    class LunchError(Exception):
        pass

    g = g1()
    for i in range(2):
        x = next(g)
        print("Yielded %s" % (x,))
    e = LunchError("tomato ejected")
    print("Throw returned", g.throw(e))
    print("Sub thrown")

    for x in g:
        print("Yielded %s" % (x,))


test_throw_catched_subgenerator_handling()


def give_cpython_generator():
    # TODO: This relies on eval not being inlined, which will become untrue.
    return eval("( x for x in range(3) )")


def gen_compiled():
    yield from give_cpython_generator()
    yield from range(7)


print(list(gen_compiled()))



# -*- coding: utf-8 -*-
#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

# All of these should be identical with correct software behavior.

print "Output with newline."
print "Output", "with", "newline."
print "Output trailing spaces ", "with ", "newline."
print "Output ",
print "with ",
print "newline."
print "Output\twith tab"
print "Output\t",
print "with tab"

# These ones gave errors with previous literal bugs:
print "changed 2"
print "foo%sbar%sfred%sbob?????"

a = "partial print"
# b doesn't exist

try:
    print a, undefined_global
except Exception, e:
    print "then occurred", repr(e)

print "No newline at the end",

x = 1
print """
New line is no soft space, is it
""", x



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Python version specifics.

This abstracts the Python version decisions. This makes decisions based on
the numbers, and attempts to give them meaningful names. Where possible it
should attempt to make run time detections.

"""

import os
import re
import sys

from nuitka.__past__ import WindowsError


def getSupportedPythonVersions():
    """Officially supported Python versions for Nuitka."""

    return ("2.6", "2.7", "3.3", "3.4", "3.5", "3.6", "3.7", "3.8", "3.9")


def getPartiallySupportedPythonVersions():
    """Partially supported Python versions for Nuitka."""

    return ()


def getSupportedPythonVersionStr():
    supported_python_versions = getSupportedPythonVersions()

    supported_python_versions_str = repr(supported_python_versions)[1:-1]
    supported_python_versions_str = re.sub(
        r"(.*),(.*)$", r"\1, or\2", supported_python_versions_str
    )

    return supported_python_versions_str


def _getPythonVersion():
    big, major, minor = sys.version_info[0:3]

    # TODO: Give up on decimal versions already.
    return big * 256 + major * 16 + min(15, minor)


python_version = _getPythonVersion()

python_version_full_str = ".".join(str(s) for s in sys.version_info[0:3])
python_version_str = ".".join(str(s) for s in sys.version_info[0:2])


def isNuitkaPython():
    """Is this our own fork of CPython named Nuitka-Python."""

    if python_version >= 0x300:
        return sys.implementation.name == "nuitkapython"
    else:
        return sys.subversion[0] == "nuitkapython"


def getErrorMessageExecWithNestedFunction():
    """Error message of the concrete Python in case an exec occurs in a
    function that takes a closure variable.
    """

    assert python_version < 0x300

    # Need to use "exec" to detect the syntax error, pylint: disable=W0122

    try:
        exec(
            """
def f():
   exec ""
   def nested():
      return closure"""
        )
    except SyntaxError as e:
        return e.message.replace("'f'", "'%s'")


def getComplexCallSequenceErrorTemplate():
    if not hasattr(getComplexCallSequenceErrorTemplate, "result"):
        try:
            # We are doing this on purpose, to get the exception.
            # pylint: disable=not-an-iterable,not-callable
            f = None
            f(*None)
        except TypeError as e:
            result = (
                e.args[0]
                .replace("NoneType object", "%s")
                .replace("NoneType", "%s")
                .replace("None ", "%s ")
            )
            getComplexCallSequenceErrorTemplate.result = result
        else:
            sys.exit("Error, cannot detect expected error message.")

    return getComplexCallSequenceErrorTemplate.result


_needs_set_literal_reverse_insertion = None


def needsSetLiteralReverseInsertion():
    """For Python3, until Python3.5 ca. the order of set literals was reversed."""
    # Cached result, pylint: disable=global-statement
    global _needs_set_literal_reverse_insertion

    if _needs_set_literal_reverse_insertion is None:
        try:
            value = eval("{1,1.0}.pop()")  # pylint: disable=eval-used
        except SyntaxError:
            _needs_set_literal_reverse_insertion = False
        else:
            _needs_set_literal_reverse_insertion = type(value) is float

    return _needs_set_literal_reverse_insertion


def needsDuplicateArgumentColOffset():
    if python_version < 0x353:
        return False
    else:
        return True


def isDebianPackagePython():
    """Is this Python from a debian package."""

    if python_version < 0x300:
        return hasattr(sys, "_multiarch")
    else:
        try:
            from distutils.dir_util import _multiarch
        except ImportError:
            return False
        else:
            return True


def isUninstalledPython():
    # Debian package.
    if isDebianPackagePython():
        return False

    if isStaticallyLinkedPython():
        return True

    if os.name == "nt":
        import ctypes.wintypes

        GetSystemDirectory = ctypes.windll.kernel32.GetSystemDirectoryW
        GetSystemDirectory.argtypes = (ctypes.wintypes.LPWSTR, ctypes.wintypes.DWORD)
        GetSystemDirectory.restype = ctypes.wintypes.DWORD

        MAX_PATH = 4096
        buf = ctypes.create_unicode_buffer(MAX_PATH)

        res = GetSystemDirectory(buf, MAX_PATH)
        assert res != 0

        system_path = os.path.normcase(buf.value)
        return not getRunningPythonDLLPath().startswith(system_path)

    return (
        os.path.exists(os.path.join(sys.prefix, "conda-meta"))
        or "WinPython" in sys.version
    )


def getRunningPythonDLLPath():
    import ctypes.wintypes

    MAX_PATH = 4096
    buf = ctypes.create_unicode_buffer(MAX_PATH)

    GetModuleFileName = ctypes.windll.kernel32.GetModuleFileNameW
    GetModuleFileName.argtypes = (
        ctypes.wintypes.HANDLE,
        ctypes.wintypes.LPWSTR,
        ctypes.wintypes.DWORD,
    )
    GetModuleFileName.restype = ctypes.wintypes.DWORD

    # We trust ctypes internals here, pylint: disable=protected-access
    res = GetModuleFileName(ctypes.pythonapi._handle, buf, MAX_PATH)
    if res == 0:
        # Windows only code, pylint: disable=I0021,undefined-variable
        raise WindowsError(
            ctypes.GetLastError(), ctypes.FormatError(ctypes.GetLastError())
        )

    dll_path = os.path.normcase(buf.value)
    assert os.path.exists(dll_path), dll_path

    return dll_path


def getTargetPythonDLLPath():
    dll_path = getRunningPythonDLLPath()

    from nuitka.Options import isPythonDebug

    if dll_path.endswith("_d.dll"):
        if not isPythonDebug():
            dll_path = dll_path[:-6] + ".dll"

        if not os.path.exists(dll_path):
            sys.exit("Error, cannot switch to non-debug Python, not installed.")

    else:
        if isPythonDebug():
            dll_path = dll_path[:-4] + "_d.dll"

        if not os.path.exists(dll_path):
            sys.exit("Error, cannot switch to debug Python, not installed.")

    return dll_path


def isStaticallyLinkedPython():
    # On Windows, there is no way to detect this from syconfig.
    if os.name == "nt":
        import ctypes

        return ctypes.pythonapi is None

    try:
        import sysconfig
    except ImportError:
        # Cannot detect this properly for Python 2.6, but we don't care much
        # about that anyway.
        return False

    result = sysconfig.get_config_var("Py_ENABLE_SHARED") == 0

    return result


def getPythonABI():
    if hasattr(sys, "abiflags"):
        abiflags = sys.abiflags

        # Cyclic dependency here.
        from nuitka.Options import isPythonDebug

        if isPythonDebug() or hasattr(sys, "getobjects"):
            if not abiflags.startswith("d"):
                abiflags = "d" + abiflags
    else:
        abiflags = ""

    return abiflags


_the_sys_prefix = None


def getSystemPrefixPath():
    """Return real sys.prefix as an absolute path breaking out of virtualenv.

    Note:

        For Nuitka, it often is OK to break out of the virtualenv, and use the
        original install. Mind you, this is not about executing anything, this is
        about building, and finding the headers to compile against that Python, we
        do not care about any site packages, and so on.

    Returns:
        str - path to system prefix
    """

    global _the_sys_prefix  # Cached result, pylint: disable=global-statement
    if _the_sys_prefix is None:

        sys_prefix = getattr(
            sys, "real_prefix", getattr(sys, "base_prefix", sys.prefix)
        )
        sys_prefix = os.path.abspath(sys_prefix)

        # Some virtualenv contain the "orig-prefix.txt" as a textual link to the
        # target, this is often on Windows with virtualenv. There are two places to
        # look for.
        for candidate in (
            "Lib/orig-prefix.txt",
            "lib/python%s/orig-prefix.txt" % python_version_str,
        ):
            candidate = os.path.join(sys_prefix, candidate)
            if os.path.exists(candidate):
                with open(candidate) as f:
                    sys_prefix = f.read()

                # Trailing spaces in the python prefix, please not.
                assert sys_prefix == sys_prefix.strip()

        # This is another for of virtualenv references:
        if os.name != "nt" and os.path.islink(os.path.join(sys_prefix, ".Python")):
            sys_prefix = os.path.normpath(
                os.path.join(os.readlink(os.path.join(sys_prefix, ".Python")), "..")
            )

        # Some virtualenv created by "venv" seem to have a different structure, where
        # library and include files are outside of it.
        if (
            os.name != "nt"
            and python_version >= 0x330
            and os.path.exists(os.path.join(sys_prefix, "bin/activate"))
        ):
            python_binary = os.path.join(sys_prefix, "bin", "python")
            python_binary = os.path.realpath(python_binary)

            sys_prefix = os.path.normpath(os.path.join(python_binary, "..", ".."))

        _the_sys_prefix = sys_prefix

    return _the_sys_prefix



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Standard plug-in to find data files.

"""

import collections
import os

from nuitka import Options
from nuitka.__past__ import basestring  # pylint: disable=I0021,redefined-builtin
from nuitka.containers.oset import OrderedSet
from nuitka.freezer.IncludedDataFiles import (
    makeIncludedDataDirectory,
    makeIncludedDataFile,
    makeIncludedEmptyDirectories,
)
from nuitka.plugins.PluginBase import NuitkaPluginBase
from nuitka.utils.FileOperations import (
    getFileList,
    resolveShellPatternToFilenames,
)


def _createEmptyDirText(filename):
    # We create the same content all the time, pylint: disable=unused-argument
    return ""


def _getSubDirectoryFolders(module, subdirs):
    """Get dirnames in given subdirs of the module.

    Notes:
        All dirnames in folders below one of the subdirs are recursively
        retrieved and returned shortened to begin with the string of subdir.
    Args:
        module: module object
        subdirs: sub folder name(s) - str or None or tuple
    Returns:
        makeIncludedEmptyDirectories of found dirnames.
    """

    module_dir = module.getCompileTimeDirectory()
    file_list = []

    if subdirs is None:
        data_dirs = [module_dir]
    elif isinstance(subdirs, basestring):
        data_dirs = [os.path.join(module_dir, subdirs)]
    else:
        data_dirs = [os.path.join(module_dir, subdir) for subdir in subdirs]

    # Gather the full file list, probably makes no sense to include bytecode files
    file_list = sum(
        (
            getFileList(
                data_dir, ignore_dirs=("__pycache__",), ignore_suffixes=(".pyc",)
            )
            for data_dir in data_dirs
        ),
        [],
    )

    if not file_list:
        msg = "No files or folders found for '%s' in subfolder(s) %r (%r)." % (
            module.getFullName(),
            subdirs,
            data_dirs,
        )
        NuitkaPluginDataFileCollector.warning(msg)

    is_package = module.isCompiledPythonPackage() or module.isUncompiledPythonPackage()

    # We need to preserve the package target path in the dist folder.
    if is_package:
        package_part = module.getFullName().asPath()
    else:
        package = module.getFullName().getPackageName()

        if package is None:
            package_part = ""
        else:
            package_part = package.asPath()

    item_set = OrderedSet()

    for f in file_list:
        target = os.path.join(package_part, os.path.relpath(f, module_dir))

        dir_name = os.path.dirname(target)
        item_set.add(dir_name)

    return makeIncludedEmptyDirectories(
        source_path=module_dir,
        dest_paths=item_set,
        reason="Subdirectories of module %s" % module.getFullName(),
    )


class NuitkaPluginDataFileCollector(NuitkaPluginBase):
    plugin_name = "data-files"

    KnownDataFileDesc = collections.namedtuple(
        "KnownDataFileDesc", ("dest_path", "filename_pattern")
    )

    known_data_files = {
        # Key is the package name to trigger it
        # Value is a tuple of 2 element tuples, thus trailing commas, where
        # the target path can be specified (None is just default, i.e. the
        # package directory) and the filename relative to the source package
        # directory
        "botocore": (KnownDataFileDesc(filename_pattern="cacert.pem", dest_path=None),),
        "site": (
            KnownDataFileDesc(filename_pattern="orig-prefix.txt", dest_path=None),
        ),
        "nose.core": (KnownDataFileDesc(filename_pattern="usage.txt", dest_path=None),),
        "scrapy": (KnownDataFileDesc(filename_pattern="VERSION", dest_path=None),),
        "dask": (KnownDataFileDesc(filename_pattern="dask.yaml", dest_path=""),),
        "cairocffi": (KnownDataFileDesc(filename_pattern="VERSION", dest_path=None),),
        "cairosvg": (KnownDataFileDesc(filename_pattern="VERSION", dest_path=None),),
        "weasyprint": (KnownDataFileDesc(filename_pattern="VERSION", dest_path=None),),
        "tinycss2": (KnownDataFileDesc(filename_pattern="VERSION", dest_path=None),),
        "certifi": (KnownDataFileDesc(filename_pattern="cacert.pem", dest_path=None),),
        "importlib_resources": (
            KnownDataFileDesc(filename_pattern="version.txt", dest_path=None),
        ),
        "moto": (
            KnownDataFileDesc(
                filename_pattern="ec2/resources/instance_types.json", dest_path=None
            ),
            KnownDataFileDesc(
                filename_pattern="ec2/resources/amis.json", dest_path=None
            ),
        ),
        "skimage": (
            # TODO: Probably should use *.ini with patterns being supported.
            KnownDataFileDesc(
                filename_pattern="io/_plugins/fits_plugin.ini", dest_path=None
            ),
            KnownDataFileDesc(
                filename_pattern="io/_plugins/gdal_plugin.ini", dest_path=None
            ),
            KnownDataFileDesc(
                filename_pattern="io/_plugins/gtk_plugin.ini", dest_path=None
            ),
            KnownDataFileDesc(
                filename_pattern="io/_plugins/imageio_plugin.ini", dest_path=None
            ),
            KnownDataFileDesc(
                filename_pattern="io/_plugins/imread_plugin.ini", dest_path=None
            ),
            KnownDataFileDesc(
                filename_pattern="io/_plugins/matplotlib_plugin.ini", dest_path=None
            ),
            KnownDataFileDesc(
                filename_pattern="io/_plugins/pil_plugin.ini", dest_path=None
            ),
            KnownDataFileDesc(
                filename_pattern="io/_plugins/qt_plugin.ini", dest_path=None
            ),
            KnownDataFileDesc(
                filename_pattern="io/_plugins/simpleitk_plugin.ini", dest_path=None
            ),
            KnownDataFileDesc(
                filename_pattern="io/_plugins/tifffile_plugin.ini", dest_path=None
            ),
        ),
        "skimage.feature._orb_descriptor_positions": (
            KnownDataFileDesc(
                filename_pattern="orb_descriptor_positions.txt", dest_path=None
            ),
        ),
        "tzdata": (KnownDataFileDesc(filename_pattern="zones", dest_path=None),),
        "lib2to3.pgen2": (
            KnownDataFileDesc(filename_pattern="../*.pickle", dest_path="lib2to3"),
        ),
    }

    # data files to be copied are contained in subfolders named as the second item
    known_data_dirs = {
        "botocore": "data",
        "boto3": "data",
        "sklearn.datasets": ("data", "descr"),
        "osgeo": "data",
        "pyphen": "dictionaries",
        "pytz": "zoneinfo",
        "pytzdata": "zoneinfo",
        "tzdata": "zoneinfo",
        "pywt": "data",
        "skimage": "data",
        "weasyprint": "css",
        "xarray": "static",
        "gooey": ("languages", "images"),
        "jsonschema": "schemas",
        "xmlschema": "schemas",
    }

    known_data_dir_structure = {
        "pendulum": "locales",
    }

    generated_data_files = {
        "Cryptodome.Util._raw_api": (
            ("Cryptodome/Util", ".keep_dir.txt", _createEmptyDirText),
        ),
        "Crypto.Util._raw_api": (
            ("Crypto/Util", ".keep_dir.txt", _createEmptyDirText),
        ),
    }

    @classmethod
    def isRelevant(cls):
        return Options.isStandaloneMode()

    @staticmethod
    def isAlwaysEnabled():
        return True

    def considerDataFiles(self, module):
        # Many cases to deal with, pylint: disable=too-many-branches

        module_name = module.getFullName()
        module_folder = module.getCompileTimeDirectory()

        if module_name in self.known_data_files:
            for know_data_file_desc in self.known_data_files[module_name]:
                target_dir = know_data_file_desc.dest_path
                filename_pattern = know_data_file_desc.filename_pattern

                if target_dir is None:
                    if (
                        module.isCompiledPythonPackage()
                        or module.isUncompiledPythonPackage()
                    ):
                        target_dir = module_name.asPath()
                    else:
                        package_name = module_name.getPackageName()

                        if package_name is not None:
                            target_dir = module_name.getPackageName().asPath()
                        else:
                            target_dir = "."

                source_path_pattern = os.path.join(module_folder, filename_pattern)

                for filename in resolveShellPatternToFilenames(source_path_pattern):
                    yield makeIncludedDataFile(
                        source_path=filename,
                        dest_path=os.path.normpath(
                            os.path.join(target_dir, os.path.basename(filename))
                        ),
                        reason="package data for %r" % module_name.asString(),
                    )

        if module_name in self.known_data_dirs:
            data_dirs = self.known_data_dirs[module_name]

            if type(data_dirs) is not tuple:
                data_dirs = (data_dirs,)

            for data_dir in data_dirs:
                source_path = os.path.join(module_folder, data_dir)

                if os.path.isdir(source_path):
                    yield makeIncludedDataDirectory(
                        source_path=source_path,
                        dest_path=os.path.join(module_name.asPath(), data_dir),
                        reason="package data directory for %r" % module_name.asString(),
                    )

        if module_name in self.known_data_dir_structure:
            empty_dirs = self.known_data_dir_structure[module_name]

            yield _getSubDirectoryFolders(module, empty_dirs)

        if module_name in self.generated_data_files:
            for target_dir, filename, func in self.generated_data_files[module_name]:
                if target_dir is None:
                    target_dir = module_name.replace(".", os.path.sep)

                yield (func, os.path.normpath(os.path.join(target_dir, filename)))



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Included data files for standalone mode.

This keeps track of data files for standalone mode. Do not should them for
DLLs or extension modules, these need to be seen by Nuitka as entry points
for dependency analysis.
"""

import collections
import os

from nuitka.utils.FileOperations import isRelativePath

IncludedDataFile = collections.namedtuple(
    "IncludedDataFile", ("kind", "source_path", "dest_path", "reason")
)


def makeIncludedEmptyDirectories(source_path, dest_paths, reason):
    for dest_path in dest_paths:
        assert not os.path.isabs(dest_path)

    # Require that to not be empty.
    assert dest_paths

    return IncludedDataFile(
        kind="empty_dirs", source_path=source_path, dest_path=dest_paths, reason=reason
    )


def makeIncludedDataFile(source_path, dest_path, reason):
    assert isRelativePath(dest_path), dest_path

    return IncludedDataFile(
        kind="data_file", source_path=source_path, dest_path=dest_path, reason=reason
    )


def makeIncludedDataDirectory(source_path, dest_path, reason):
    assert isRelativePath(dest_path), dest_path

    return IncludedDataFile(
        kind="data_dir", source_path=source_path, dest_path=dest_path, reason=reason
    )



#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Progress bars in Nuitka.

This is responsible for wrapping the rendering of progress bar and emitting tracing
to the user while it's being displayed.

"""

from nuitka import Tracing
from nuitka.Tracing import general
from nuitka.utils.Importing import importFromInlineCopy
from nuitka.utils.ThreadedExecutor import RLock
from nuitka.utils.Utils import isWin32Windows

# Late import and optional to be there.
use_progress_bar = False
tqdm = None


class NuitkaProgessBar(object):
    def __init__(self, stage, total, min_total, unit):
        self.stage = stage
        self.total = total

        # The minimum may not be provided, then default to 0.
        self.min_total = min_total or 0
        self.unit = unit

        # No item under work yet.
        self.item = None

        # No progress yet.
        self.progress = 0

        # Render immediately with 0 progress.
        self._reinit()

    def _reinit(self):
        # Note: Setting disable=None enables tty detection.
        self.tqdm = tqdm(
            initial=self.progress,
            total=max(self.total, self.min_total),
            unit=self.unit,
            disable=None,
            leave=False,
            bar_format="{desc}{percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt}{postfix}",
        )

        self.tqdm.set_description(self.stage)
        self.setCurrent(self.item)

    def updateTotal(self, total):
        if total != self.total:
            self.total = total
            self.tqdm.total = max(total, self.min_total)

    def setCurrent(self, item):
        if item != self.item:
            self.item = item

            if item is not None:
                self.tqdm.set_postfix_str(item)
            else:
                self.tqdm.set_postfix()

    def update(self):
        self.progress += 1
        self.tqdm.update(1)

    def clear(self):
        self.tqdm.clear()

    def hideProgressBar(self):
        # TODO: Need a better way to do this.
        self.tqdm.clear()

    def resumeProgressBar(self):
        self._reinit()

    def close(self):
        self.tqdm.close()


def enableProgressBar():
    global use_progress_bar  # singleton, pylint: disable=global-statement
    global tqdm  # singleton, pylint: disable=global-statement

    if isWin32Windows():
        colorama = importFromInlineCopy("colorama", must_exist=True)
        colorama.init()

    tqdm = importFromInlineCopy("tqdm", must_exist=False)

    if tqdm is None:
        try:
            # Cannot use import tqdm due to pylint bug.
            import tqdm as tqdm_installed  # pylint: disable=I0021,import-error

            tqdm = tqdm_installed
        except ImportError:
            # We handle the case without inline copy too, but it may be removed, e.g. on
            # Debian it's only a recommended install, and not included that way.
            pass

    # Tolerate the absence ignore the progress bar
    if tqdm is not None:
        tqdm = tqdm.tqdm

        tqdm.set_lock(RLock())
        use_progress_bar = True


def setupProgressBar(stage, unit, total, min_total=0):
    # Make sure the other was closed.
    assert Tracing.progress is None

    if use_progress_bar:
        Tracing.progress = NuitkaProgessBar(
            stage=stage,
            total=total,
            min_total=min_total,
            unit=unit,
        )


def reportProgressBar(item, total=None, update=True):
    if Tracing.progress is not None:
        try:
            if total is not None:
                Tracing.progress.updateTotal(total)

            Tracing.progress.setCurrent(item)

            if update:
                Tracing.progress.update()
        except Exception:  # Catch all the things, pylint: disable=broad-except
            # We disable the progress bar now, because it's causing issues.
            general.warning("Progress bar disabled due to bug")
            closeProgressBar()


def closeProgressBar():
    """Close the active progress bar.

    Returns: int or None - if displayed, the total used last time.
    """

    if Tracing.progress is not None:
        # Retrieve that previous total, for repeated progress bars, it
        # can be used as a new minimum.
        result = Tracing.progress.total

        Tracing.progress.close()
        Tracing.progress = None

        return result
